#define INCL_BASE
#include <os2.h>
#include <stdio.h>
#include <syscall.h>


#define MMIO_CONFIGURATION_BASE_ADDRESS  0xC0010058
#define ONE_MEGABYTE                     1024UL*1024UL*1UL

#define DEVICE                           0x18
#define FUNCTION                         0x3
#define TEMPCTRL                         0xA4
#define NORTHBRIDGECAPS                  0xE8

#define MMIOADDRESS(bus,device,function,offset) (((1UL*(bus)) << 20) + ((1UL*(device)) << 15) + ((1UL*(function)) << 12) + (offset))

typedef struct
{
   ULONG ulStepping:4;
   ULONG ulModel:4;
   ULONG ulFamily:4;
   ULONG ulProcessorType:2;
   ULONG ulReserved:2;
   ULONG ulExtendedModel:4;
   ULONG ulExtendedFamily:8;
} VERSIONINFO;

typedef struct
{
   ULONG ulBrand:8;
   ULONG ulCLFLUSHlineSize:8;
   ULONG ulNumLogicalProcessors:8;
   ULONG ulInitialAPICID:8;
} ADDCPUINFO, *PADDCPUINFO;

typedef struct
{
   VERSIONINFO bfVersionInfo;
   ADDCPUINFO bfAddInfo;
   ULONG ulBaseFeatures;
   ULONG ulExtendedFeatures;
} CPUFEATURES, *PCPUFEATURES;


typedef struct {

   union {
     struct {
       ULONG PerStepTimeUp:5;
       ULONG TmpMaxDiffUp:2;
       ULONG TmpSlewDnEn:1;
       ULONG PerStepTimeDn:5;
       ULONG res1:3;
       ULONG CurTmpTjSel:2;
       ULONG res2:3;
       ULONG CurTmp:11;
     } bits;
     ULONG dword;
   } u;
} REPTMPCTRL, *PREPTMPCTRL;



/*
 * AllocFlags
 */
#define VMDHA_16MB        0x001 /* (0000 0000 00001B)
                                 * If set, the object will be allocated below
                                 * the 16M line
                                 */
#define VMDHA_FIXED       0x002 /* (0000 0000 00010B)
                                 * If set, the object must be in fixed memory
                                 * at all times.
                                 * If clear, then the object may be moved or
                                 * paged out as is necessary.
                                 */
#define VMDHA_SWAP        0x004 /* (0000 0000 0100B)
                                 * If set, swappable memory will be allocated
                                 * for the object.
                                 * If clear, the memory will be movable or
                                 * fixed dependending on the setting of
                                 * VMDHA_FIXED.
                                 * If set, VMDHA_FIXED must be clear.
                                 */
#define VMDHA_CONTIG      0x008 /* (0000 0000 1000B)
                                 * If set, the object needs to be allocated in
                                 * contiguous memory.
                                 * If clear, the the pages may be discontiguous
                                 * in physical memory.
                                 * In order to request contiguous memory, the
                                 * PDD must have also requested fixed memory.
                                 * (VMDHA_FIXED must be set)
                                 */
#define VMDHA_PHYS        0x010 /* (0000 0001 0000B)
                                 * If set, a linear address mapping will be
                                 * obtained for the physical passed in the
                                 * PhysAddr pointer.
                                 */
#define VMDHA_PROCESS     0x020 /* (0000 0010 0000B)
                                 * If set, the linear address returned will be
                                 * in the process address range.
                                 * If clear, the allocation will be done in the
                                 * global address range, that is, accessible
                                 * outside of the current process' context.
                                 */
#define VMDHA_SGSCONT     0x040 /* (0000 0100 0000B)
                                 * If set, the allocated memory can be
                                 * registered under screen group switch control.
                                 * This flag is valid only if mapping is in
                                 * process address range.  (VMDHA_PROCESS must
                                 * be set)
                                 */
#define VMDHA_SELMAP      0x080 /* (0000 1000 0000B)
                                 * If set, provide a selector mapping for the
                                 * linear address range.
                                 */
#define VMDHA_RESERVE     0x100 /* (0001 0000 0000B)
                                 * If set, the memory will only be reserved.
                                 * No committment will be done and any attempt
                                 * to access reserved but not committed memory
                                 * will cause a fatal page fault.
                                 */
#define VMDHIA_USER       0x200 /* (0010 0000 0000B)
                                 * If set, provides user mode access.
                                 * (init time only)
                                 */
#define VMDHA_SHARED      0x400 /* (0100 0000 0000B)
                                 * If set, Allocate shared memory object. This
                                 * flag is only supported for mapping a
                                 * physical address to non-system memory into
                                 * the shared arena. VMDHA_PHYS must be
                                 * specified.
                                 */
#define VMDHA_USEHIGHMEM  0x800 /* (1000 0000 0000B)
                                 * If set, the object will be allocated above
                                 * the 16MB line if possible.  If memory above
                                 * 16MB exists but there is not enough to
                                 * satisfy the request, the memory above 16MB
                                 * will be used first and the remainder taken
                                 * from below 16MB.  If no memory above 16MB
                                 * exists, the allocation will be taken from
                                 * existing memory.
                                 * This bit is valid only during device driver
                                 * initialization.  If this bit is used at any
                                 * other time, KernVMAlloc will return an error.
                                 * All other bits must be clear.
                                 */
#define VMDHA_ALIGN64K 0x1000   /* (0001 0000 0000 0000B)
                                 * If set, the linear address of the object
                                 * will be aligned on 64K boundary.
                                 */

#define VMDHA_USEHMA   0x2000   /* (0010 0000 0000 0000B)
                                 * If set, the linear address of the object
                                 * will be allocated in the HMA
                                 */


APIRET APIENTRY KernVMAlloc (ULONG   cBytes,
                             ULONG   VMAllocFlags,
                             VOID  **ppLinAddr,
                             VOID  **ppPhysAddr,
                             VOID  **SysReserved);

APIRET  APIENTRY KernVMFree (VOID *pLinAddr);

extern ULONGLONG APIENTRY ReadMSR(ULONG msrNum);
extern BOOL APIENTRY IsAMD(VOID);
extern APIRET APIENTRY GetCPUFeatures(PCPUFEATURES p);



APIRET APIENTRY QueryConfigBase(ULONG arg1,ULONG arg2) /* will return the MMIO Config Space physical address as can be found in the MSR */
{
   APIRET rc = NO_ERROR;
   PULONGLONG pullResult = (PULONGLONG)arg1;
   *pullResult = ReadMSR(MMIO_CONFIGURATION_BASE_ADDRESS);
   if (!(pullResult->ulLo & 0x1)) {
      rc = ERROR_INVALID_ADDRESS;
   } else {
      pullResult->ulLo &= ~0x3F;
      rc = NO_ERROR;
   } /* endif */
   return rc;
}

APIRET APIENTRY MapMMIOConfigSpace(ULONG arg1,ULONG arg2) /* maps the MMIO Config space physical address to a linear address usable in current process */
{
   APIRET rc = KernVMAlloc(ONE_MEGABYTE*256,VMDHA_PHYS | VMDHA_PROCESS | VMDHA_USEHMA,(PVOID *)arg2,(PVOID *)arg1,NULL);
   return 0;
}

APIRET APIENTRY FreeLinearMapping(ULONG arg1,ULONG arg2) /* undoes the physical to linear mapping */
{
   APIRET rc = KernVMFree((PVOID)arg1);
   return rc;
}

int main(int argc,char *argv[])
{
   USERPFN FuncTable[] = {QueryConfigBase,MapMMIOConfigSpace,FreeLinearMapping};
   WRAPPERPFN pfnWrapper = NULL;
   ULONGLONG ullResult = {0};
   APIRET rc = NO_ERROR;
   ULONG pPhysMMIOSpaceAddr = 0;
   PVOID pLinMMIOSpaceAddr = NULL;
   PULONG pTempReg = NULL;
   REPTMPCTRL ulVal = {0};
   CPUFEATURES cpuFeatures = {0};
   ULONG ulModel = 0;
   ULONG ulFamily = 0;
   ULONG ulLocalAPIC = 0;
   MPAFFINITY affinity = {0x3,0};

   rc = DosSetThreadAffinity(&affinity);

   if (!IsAMD()) {
      printf("This util only supports AMD CPUs, quitting...\n");
      return 1;
   } /* endif */

   GetCPUFeatures(&cpuFeatures);
   ulModel = cpuFeatures.bfVersionInfo.ulModel;
   if (ulModel >= 0xF) {
      ulModel += cpuFeatures.bfVersionInfo.ulExtendedModel;
   } /* endif */
   ulFamily = cpuFeatures.bfVersionInfo.ulFamily;
   if (ulFamily >= 0xF) {
      ulFamily += cpuFeatures.bfVersionInfo.ulExtendedFamily;
   } /* endif */
   ulLocalAPIC = cpuFeatures.bfAddInfo.ulInitialAPICID;

   printf("AMD CPU: Family:0x%.2x, Model:0x%.2x, Local APIC:0x%.2x\n",ulFamily,ulModel,ulLocalAPIC);

   pfnWrapper = registerUserFunctions(FuncTable,sizeof(FuncTable)/sizeof(FuncTable[0]));
   if (pfnWrapper) {
      rc = pfnWrapper(0,(ULONG)&ullResult,0); /* call QueryConfigBase to get the physical base of MMIO Space */
      if (NO_ERROR == rc) {
        printf("MMIO Configuration Base Address:0x%.8X%.8X\n",ullResult.ulHi,ullResult.ulLo);

        pPhysMMIOSpaceAddr = ullResult.ulLo;
        rc = pfnWrapper(1,(ULONG)&pPhysMMIOSpaceAddr,(ULONG)&pLinMMIOSpaceAddr); /* call MapMMIOConfigSpace to get a linear mapping into the current process */

        pTempReg = (PULONG)((ULONG)pLinMMIOSpaceAddr + MMIOADDRESS(0,DEVICE,FUNCTION,TEMPCTRL));

#if 0
        /* read current settings */
        ulVal.u.dword = *pTempReg;
        ulVal.u.bits.TmpSlewDnEn = 0; // do not allow to update Tctl to follow if measured temperature drops below Tctl
        ulVal.u.bits.TmpMaxDiffUp = 2;// allow Tctl to follow if measured temperature climbs by 3 C above Tctl
        ulVal.u.bits.PerStepTimeDn = 0x18;
        ulVal.u.bits.PerStepTimeUp = 0x18;
        /* do it, set new values */
        *pTempReg = ulVal.u.dword;
#endif

        /* do it, read the register to query Ttl */
        ulVal.u.dword = *pTempReg;
        printf("Temperature core 0 (0x%.8X):%.2f C\n",ulVal.u.dword,ulVal.u.bits.CurTmp * 0.125);

        rc = pfnWrapper(2,(ULONG)pLinMMIOSpaceAddr,0); /* free linear mapping */
      }
         else {
         printf("Could not return the SYSCALL wrapper function address.\n");
      } /* endif */
   } /* endif */
   return 0;
}
